open Generator
  
type test = {
    weighing : weighing;
    mutable entropy : float;
  }

      
type result =
  | Left
  | Right
  | Equilibrium


(* cre tous les tests possibles avec card pices <= max_coin *)
let generate_tests card max_coin =

  (* construit un test  partir d'une pese *)
  let build_test w = { weighing = w;
		       entropy = 0.;
		     }
  in

  List.map build_test (generate_all_weighings card max_coin);;


(* gnre la distribution de probabilits initiale *)
(* p0 vaut all_true et les autres sont uniformes *)
let generate_probabilities all_true card =

  let n = 2*card + 1 in
  let p = (1. -. all_true) /. (float_of_int (n-1)) in
  let result = Array.make n p in
  result.(0) <- all_true;
  result;;

let ln x =
  if x = 0.
  then 0.
  else (log x) /. (log 3.);;
    
(* trouve le test avec la plus petite entropie dans la liste *)
let rec find_min_entropy_test test_list best =
  match test_list with
  | [] -> best
  | first::tail ->
      match best with
      |	None -> find_min_entropy_test tail (Some first)
      |	Some t ->
	  if first.entropy<t.entropy
	  then find_min_entropy_test tail (Some first)
	  else find_min_entropy_test tail (Some t);;
	      

(* renvoie les probabilits des rsultats d'une pese sous la forme *)
(* p(gauche),p(droite),p(quilibre) *)
let probabilities test p_value =
  (* nombre de pices *)
  let n = ((Array.length p_value)-1)/2 in
  (* pese du test *)
  let w = test.weighing in
  (* pices  gauche et  droite *)
  let left = fst w and right = snd w in
  (* probabilit de pencher  gauche et  droite *)
  let left_p = ref 0. and right_p = ref 0. in
  for i = 1 to n do
    if in_set i left
    then
      begin
	left_p := !left_p +. p_value.(i+n);
	right_p := !right_p +. p_value.(i);
      end
    else
      if in_set i right
      then
	begin
	  left_p := !left_p +. p_value.(i);
	  right_p := !right_p +. p_value.(i+n);
	end
  done;
  !left_p,!right_p,(1. -. (!left_p +. !right_p));;


(* calcule l'entropie associe  une distribution de probabilit *)
let entropy p_value =
  let n = Array.length p_value in
  let result = ref 0. in
  for i = 0 to n-1 do
    let p = p_value.(i) in
    result := !result -. (p *. ln p)
  done;
  !result;;


(* renvoie la distribution de probabilits conditionnelles si test aboutit *)
(* au rsultat hypothese *)    
let learn test hypothese p_value =
  (* nombre de pices *)
  let n = ((Array.length p_value)-1)/2 in
  (* pese du test *)
  let w = test.weighing in
  (* pices  gauche et  droite *)
  let left = fst w and right = snd w in
  (* nouvelles probabilits si l'hypothse se ralise *)
  let result = Array.copy p_value in
  begin
    match hypothese with
    | Left ->
	begin
  	  (* on n'est pas dans le cas sans fausse pice *)
	  result.(0) <- 0.;
	  for i = 1 to n do
	    (* si la pice est  gauche quand la balance penche  gauche *)
	    if in_set i left
	    then
	      (* la pice n'est pas trop lgre *)
	      result.(i) <- 0.
	    else
	      (* si elle est  droite *)
	      if in_set i right
	      then
	        (* elle n'est pas trop lourde *)
		result.(i+n) <- 0.
	      else
		begin
		  (* si elle n'est pas dans la pese, elle est vraie *)
		  result.(i) <- 0.;
		  result.(i+n) <- 0.;   
		end
	  done;
	end
    | Right ->
	begin
	  (* on n'est pas dans le cas sans fausse pice *)
	  result.(0) <- 0.;
	  for i = 1 to n do
	    (* si la pice est  gauche quand la balance penche  droite *)
	    if in_set i left
	    then
	      (* la pice n'est pas trop lourde *)
	      result.(i+n) <- 0.
	    else
	      (* si elle est  droite *)
	      if in_set i right
	      then
	        (* elle n'est pas trop lgre *)
		result.(i) <- 0.
	      else
		begin
		  (* si elle n'est pas dans la pese, elle est vraie *)
		  result.(i) <- 0.;
		  result.(i+n) <- 0.;   
	        end
	  done;
	end
    | Equilibrium ->
	for i = 1 to n do
	  if (in_set i left)||(in_set i right)
	  then
	    begin
	      (* si la balance ne penche pas, les pices peses sont vraies *)
	      result.(i) <- 0.;
	      result.(i+n) <- 0.;
	    end
	done;
  end;
  (* liste des probabilits sauf la premire (on ne compte pas le cas *)
  (* sans fausse pice) *)
  let p_list = List.tl (Array.to_list result) in
  (* nombre de cas encore indcis *)
  let nb_unknown = List.length (List.filter (function x -> (x<>0.)) p_list) in
  (* s'il y a encore des pices indcises *)
  if nb_unknown <> 0
  then
    (* nouvelle probabilit pour les cas indcis *)
    let new_p = (1. -. result.(0)) /. float_of_int(nb_unknown) in
    (* mise  jour des probabilits *)
    for i = 1 to 2*n do
      if result.(i) <> 0.
      then result.(i) <- new_p
    done;
  (* s'il n'y a plus de pices indcises, il n'y pas de contrefaon *)
  else
    result.(0) <- 1.;
  result;;
  

(* met  jour le champ entropy de tous les tests *)
let update_all_tests p_value test_list =

  let update_test p_value test =
    let (left_p,right_p,equi_p) = probabilities test p_value in
    let (left_e,right_e,equi_e) = (entropy (learn test Left p_value),
				   entropy (learn test Right p_value),
				   entropy (learn test Equilibrium p_value)) in
    test.entropy <- (left_p *. left_e)+.(right_p *. right_e)+.(equi_p *. equi_e)
  in
  List.iter (update_test p_value) test_list;;
    
    












