open Generator
open Entropy
open Graphics

  
type tree =
  | Node of weighing*tree*tree*tree
  | Leaf of coin option
  | Impossible


type huffmann_element = {
    proba : float;
    elt : tree;
  } 
      

(* fonction debug : affiche la liste des probabilits*)      
let print_prob p_value =
  let n = Array.length p_value in
  for i = 0 to n-1 do
    print_float p_value.(i);
    print_string " ";
  done;
  print_newline ();;


(* construit l'arbre des tests  effectuer *)
let rec build_tree p_value test_list =
  (* nombre de cas  envisager *)
  let n = Array.length p_value in
  (* entropie de la situation actuelle *)
  let e = entropy p_value in
  (* si l'entropie est nulle, on a trouv *)
  if e = 0.
  then
    let l = ref 0 in
    begin
      (* on cherche s'il y a une pice dont on est sr qu'elle est fausse *)
      for i = 1 to n-1 do
	if p_value.(i) <> 0.
	then l := i
      done;
      if !l = 0
      (* si on ne trouve pas, il n'y a pas de fausse pice *)	  
      then Leaf None
      (* sinon on donne la fausse pice *)
      else Leaf (Some !l)
    end
  (* si l'entropie est non nulle, on continue  construire l'arbre *)    
  else
    begin
      (* on met  jour l'entropie de tous les tests *)
      update_all_tests p_value test_list;
      let grab_test = function
	| Some t -> t
	| None -> failwith "rien dans l'option"
      in
      (* on choisit le test qui fait le plus baisser l'entropie *)
      let best = grab_test (find_min_entropy_test test_list None) in
      (* on calcule les probabilits de chaque issue pour ce test *)
      let (left_p,right_p,equi_p) = probabilities best p_value in
      (* pour une issue donne, soit construit l'arbre des tests  suivre *)
      (* soit indique que le cas n'a pas  tre envisag car incompatible *)
      (* avec les rsultats prcdents *)
      let tree_or_impossible p r =
	if p<>0.
	then build_tree (learn best r p_value) test_list
	else Impossible
      in
      let l_tree = tree_or_impossible left_p Left
      and r_tree = tree_or_impossible right_p Right 
      and e_tree = tree_or_impossible equi_p Equilibrium
      in
      Node (best.weighing,
	    l_tree,
	    r_tree,
	    e_tree)
    end


(* construit l'arbre de Huffmann correspondant  la rpartition de probabilit *) 
let build_huffmann_tree p_value =

  let init_list =
    let generate_h_elt c p = { proba = p;
			       elt = Leaf (Some c); }
    in Array.to_list (Array.mapi generate_h_elt p_value)
  in

  let rec merge_h_elt = function
    | [] -> failwith "aucun lment"
    | a::[] -> a::[]
    | a::b::[] -> { proba = a.proba +. b.proba;
		    elt = Node(default_weighing,a.elt,b.elt,Impossible);}::[]
    | h_list ->
	let h_cmp e1 e2 =
	  if e1.proba<e2.proba
	  then -1
	  else
	    if e1.proba>e2.proba
	    then 1
	    else 0
	in
	let s_list = List.sort h_cmp h_list in
	if (List.length s_list) mod 2 = 1
	then
	  match s_list with
	  | a::b::c::tail ->
	      merge_h_elt ({ proba = a.proba +. b.proba +. c.proba;
			     elt = Node(default_weighing,a.elt,b.elt,c.elt);}::tail)
	  | _ -> failwith "ne doit pas arriver"
	else
	  match s_list with
	  | a::b::tail ->
	      merge_h_elt ({ proba = a.proba +. b.proba;
			     elt = Node(default_weighing,a.elt,b.elt,Impossible);}::tail)
	  | _ -> failwith "ne doit pas arriver"
  in
  match (merge_h_elt init_list) with
  | result::[] -> result.elt
  | _ -> failwith "problme grave dans merge_h_elt"

      
(* renvoie la profondeur d'un arbre *)
let rec depth t =

  (* renvoie le maximum de trois nombres *)
  let max a b c =
    if (a>b)&&(a>c)
    then a
    else
      if b>c
      then b
      else c
  in
  match t with
  | Node(_,l,r,e) -> 1 + max (depth l) (depth r) (depth e)
  | _ -> 0;;


(* renvoie la profondeur moyenne de l'arbre *)
let mean_depth t p_value =

  let result = ref 0. in

  let rec explore depth = function
    | Node (_,l,r,e) ->
	begin
	  explore (depth+1) l;
	  explore (depth+1) r;
	  explore (depth+1) e;
	end;
    | Leaf (Some c) -> result := !result +. p_value.(c)*.(float_of_int depth);
    | Leaf None -> result := !result +. p_value.(0)*.(float_of_int depth);
    | _ -> ();
  in

  explore 0 t;
  !result;;
  
(* sortie texte d'un arbre *)
let rec print_tree indent = function
  | Node(w,l,r,e) ->	
      let new_i = String.concat "" ("**"::indent::[]) in
      begin
	print_tree new_i l;
	print_tree new_i r;
	print_tree new_i e;
      end
  | Leaf (Some c) ->
      begin
	print_string indent;
	print_int c;
	print_newline ();
      end
  | Leaf None ->
      begin
	print_string indent;
	print_string "no counterfeit";
	print_newline ();
      end
  | Impossible ->
      begin
	print_string indent;
	print_string "impossible";
	print_newline ();
      end;;

	
(* sortie graphique d'un arbre *)
let show_tree t max_coin =
  let d = depth t in
  let delta_h = 100 in

  let rec x_size tr =
    let unit_width = 35 in
    match tr with
    | Node(_,l,r,e) -> (x_size l) + (x_size r) + (x_size e)
    | _ -> unit_width
  in
  (*35*(int_of_float (3.**(float_of_int d)))*)
  let width = (x_size t)
  and height = (d+1)*delta_h + 10 in
  let sw = string_of_int width and sh = string_of_int height in
  let display = String.concat "" (" "::sw::"x"::sh::"+10+10"::[]) in
  begin
    open_graph display;
    let write s x y =
      let (tx,ty) = text_size s in
      let delta = 5 in
      let (wx,wy) = (x - tx/2,y - ty - delta) in
      begin
	draw_rect (wx - delta) (wy - delta) (tx + 2*delta) (ty + 2*delta);
	moveto wx wy;
	draw_string s;
	moveto x (wy - delta - 1);
      end
    in

    let rec draw_tree x_min x_max y = function
      |	Node (w,l,r,e) ->
	  let left = fst w and right = snd w in
	  let llist = List.map string_of_int (elements left)
	  and rlist = List.map string_of_int (elements right) in
	  let s = String.concat " " (llist@("&"::[])@rlist) in
	  begin
	    set_color black;
	    write s ((x_min+x_max)/2) y;
	    let (ox,oy) = current_point ()
	    and ny = y - delta_h
	    and wl = x_size l
	    and wr = x_size r
	    and we = x_size e in
	    begin
	      lineto (x_min + wl/2) ny;
	      moveto ox oy;
	      lineto (x_min + wl + we/2) ny;
	      moveto ox oy;
	      lineto (x_min + wl + we + wr/2) ny;
	      draw_tree x_min (x_min + wl - 1) ny l;
	      draw_tree (x_min + wl) (x_min + wl + we -1) ny e;
	      draw_tree (x_min + wl + we - 1) (x_min + wl + we + wr - 1) ny r;
	    end
	  end
      |	Leaf (Some 0) -> (* pour l'affichage d'un arbre de Huffmann uniquement *)
	  begin
	    set_color blue;
	    write "-" ((x_min+x_max)/2) y;
	  end
      |	Leaf (Some c) ->
	  begin
	    set_color blue;
	    let s =
	      if c>max_coin
	      then String.concat "" ((string_of_int (c-max_coin))::"+"::[])
	      else String.concat "" ((string_of_int c)::"-"::[])
	    in
	    write s ((x_min+x_max)/2) y;
	  end
      |	Leaf None ->
	  begin
	    set_color blue;
	    write "-" ((x_min+x_max)/2) y;
	  end
      |	Impossible ->
	  begin
	    set_color red;
	    write "X" ((x_min+x_max)/2) y;
	  end
    in

    draw_tree 0 (width-1) (height-10) t;
    while read_key  () <>' ' do () done;
  end;;
	







